import graphene

from gdaps.api import Interface


@Interface
class IGrapheneSchema:
    """Interface class to collect all graphene queries/mutations

    Any GDAPS plugin that exposes data to the GraphQL API must implement this
    Interface. Have a look at
    http://docs.graphene-python.org/projects/django/en/latest/tutorial-plain/#hello-graphql-schema-and-object-types
    how to create abstract Graphene query objects. You just need to subclass IGrapheneObject,
    and they are included into the global GraphQL API automatically.
    """

    __service__ = False
    query: type(graphene.ObjectType) = None
    mutation: type(graphene.ObjectType) = None



from copy import copy

from django import template
from django.template import Template
from django.template.context import Context

from ..api import InterfaceRegistry, Interface
from ..api.interfaces import ITemplatePluginMixin

register = template.Library()


@register.tag(name="render_plugins")
def do_render_plugins(parser, token):
    """A template tag that renders all plugins that implement the given interface."""
    try:
        tag_name, interface_name = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires exactly one argument: The interface to render"
            % token.contents.split()[0]
        )
    # remove quotes, if there
    if interface_name[0] == interface_name[-1] and interface_name[0] in ('"', "'"):
        interface_name = interface_name[1:-1]
    if not interface_name.isidentifier():
        raise template.TemplateSyntaxError(
            f"'{interface_name}' is not a valid Python identifier."
        )
    return PluginNode(InterfaceRegistry.get(interface_name))


class PluginNode(template.Node):
    def __init__(self, interface: Interface):
        self.interface = interface

    def render(self, context: Context) -> str:
        if self not in context.render_context:
            context.render_context[self] = self.interface

        interface = context.render_context[self]
        content = ""
        for plugin in interface:  # type: ITemplatePluginMixin
            # TODO is copy() needed? Test thread-safety!
            plugin_context = plugin.get_plugin_context(context)
            if plugin.template:
                template = Template(plugin.template)
            else:
                template = plugin_context.template.engine.get_template(
                    plugin.template_name
                )
            content += template.render(context=plugin_context)
        return content



import logging

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.db import connections
from django.db.migrations.executor import MigrationExecutor
from django.utils.translation import gettext_lazy as _

from gdaps.exceptions import PluginError
from gdaps.models import GdapsPlugin
from gdaps.pluginmanager import PluginManager
from semantic_version import Version

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """This is the management command to sync all installed plugins into the database.

    It parses all installed plugins, and creates shadowed meta data in the database as
    "GdapsPlugin" objects. If there is an object in the DB which was deleted on disk, it
    removes the DB object too.
    """

    # TODO: update procedure of apps

    help = "Synchronizes all installed plugins into the database."
    verbosity = 0

    __db_synchronized = False

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            nargs="?",
            type=str,
            help="synchronizes only plugins from specific database",
        )

    def is_database_synchronized(self, database=None):
        # cached flag if db is in sync, taken from migrate mgmt cmd
        if not database:
            database = "default"
        if not self.__db_synchronized:
            connection = connections[database]
            connection.prepare_database()
            executor = MigrationExecutor(connection)
            targets = executor.loader.graph.leaf_nodes()
            self.__db_synchronized = False if executor.migration_plan(targets) else True

        return self.__db_synchronized

    @staticmethod
    def _copy_plugin_to_db(app: AppConfig, db_plugin: GdapsPlugin) -> None:
        meta = app.PluginMeta
        db_plugin.name = app.name
        db_plugin.verbose_name = getattr(
            meta, "verbose_name", app.name.replace("_", " ").capitalize()
        )
        db_plugin.author = getattr(meta, "author", _("unknown"))
        db_plugin.author_email = getattr(meta, "author_email", "")
        db_plugin.vendor = getattr(meta, "vendor", "")
        db_plugin.category = getattr(meta, "category", _("Miscellaneous"))
        db_plugin.description = getattr(meta, "description", "")
        db_plugin.hidden = getattr(meta, "hidden", False)
        db_plugin.version = app.PluginMeta.version
        db_plugin.compatibility = getattr(meta, "compatibility", "")

        db_plugin.save()

    def handle(self, *args, **options) -> None:
        """Synchronizes all found plugins into the database.

        Only plugins that are activated vie INSTALLED_APPS or installed via pip and found by the
        PluginManager are taken into account.
        """
        self.verbosity = options["verbosity"]
        if not options["database"]:
            options["database"] = "default"

        logger.info(" ⌛ Searching for plugins...")
        for app in PluginManager.plugins():
            logger.info(f"   ➤ {app.name} ({app.PluginMeta.version})")

            try:
                # first, try to fetch this plugin from the DB - if doesn't exist, create and initialize it.
                # if it exists, check if there is an update available.
                # noinspection PyUnresolvedReferences
                db_plugin = GdapsPlugin.objects.get(name=app.name)

                file_version = app.PluginMeta.version
                if Version(file_version) > Version(db_plugin.version):
                    # there is a newer version available on disk

                    # TODO: check PluginMeta.compatibility here

                    logger.warning(
                        f"     There is a newer version of the '{app.verbose_name}' plugin available."
                    )

                    for database in settings.DATABASES.keys():
                        if not self.is_database_synchronized(database):
                            raise PluginError(
                                "Plugin version upgrade detected. Please run the 'migrate' management "
                                "command first to update plugins."
                            )

                # at this point, the db is in sync with the files according to Django.
                # we can now update all db fields with that from the plugin on disk
                # this is necessary as the author e.g. could change during plugin development

                # Copy metadata to DB in any case, as other things as version could have changed too.
                self._copy_plugin_to_db(app, db_plugin)

                # we can now check if there is code waiting to execute.
                # TODO: run upgrade procedure of plugin

            except ObjectDoesNotExist:
                # if it doesn't exist, it is a new plugin.
                # Let's initialize it.
                logger.info(f"     Plugin '{app.verbose_name}' is new, syncing to DB.")
                db_plugin = GdapsPlugin()
                version = None
                try:
                    version = app.PluginMeta.version
                    v = Version(version)
                    db_plugin.version = version
                except ValueError:
                    raise ImproperlyConfigured(
                        f"Plugin '{app.name}' version number is incorrect: '{version}'"
                    )
                self._copy_plugin_to_db(app, db_plugin)

                # TODO: add compatibility check

                if hasattr(app.PluginMeta, "install"):
                    try:
                        app.PluginMeta.install()
                    except Exception as E:
                        raise PluginError(
                            f"Error calling install() method of '{app.name}' plugin"
                        )

        # are there plugins in the database that do not exist on disk?
        logger.info(" ⌛ Searching for orphaned plugins in database...")

        for plugin in PluginManager.orphaned_plugins():  # type: GdapsPlugin
            if self.is_database_synchronized(options["database"] or None):
                plugin.delete()
                logger.info(f"   ➤ {plugin.name} removed from database.")
            else:
                logger.info(f"   ➤ {plugin.name}")



__all__ = ["PluginError", "IncompatibleVersionsError"]


class PluginError(Exception):
    """An Exception that marks an error in a plugin specific setting."""


class IncompatibleVersionsError(PluginError):
    """Exception that occurs when plugins that are not compatible are installed together."""



from django.contrib import admin

from gdaps.conf import gdaps_settings
from gdaps.models import GdapsPlugin


class ReadOnlyAdmin(admin.ModelAdmin):
    """This is a ModelAdmin class that sets the whole GDAPS model readonly."""

    def get_readonly_fields(self, request, obj=None):
        return [f.name for f in obj._meta.fields]

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False


class GdapsInline(admin.TabularInline):
    pass


class GdapsAdmin(ReadOnlyAdmin):
    """ModelAdmin class for GDAPS plugins."""

    list_display = [
        "verbose_name",
        "version",
        "author",
        "hidden",
        "enabled",
        "category",
    ]
    sortable_by = ["category", "hidden", "enabled"]

    def get_readonly_fields(self, request, obj=None):
        return [
            name
            for name in super().get_readonly_fields(request, obj)
            if not name in ["enabled", "hidden"]
        ]


# show Admin per default, disable this behaviour using:
# GDAPS["ADMIN"] = False
# in settings.py
if gdaps_settings.ADMIN:
    admin.site.register(GdapsPlugin, GdapsAdmin)



# Generated by Django 4.1.3 on 2022-11-20 22:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("gdaps", "0002_alter_gdapsplugin_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="gdapsplugin",
            name="visible",
        ),
        migrations.AddField(
            model_name="gdapsplugin",
            name="hidden",
            field=models.BooleanField(default=False),
        ),
    ]



from django.template.context import Context


class InterfaceNotFound(Exception):
    """Interface with this name was not found"""

    pass


class ITemplatePluginMixin:
    """A mixin that can be inherited from to build renderable plugins.

    Create an interface that inherits from ITemplatePluginMixin. Each
    implementation must either have a `template` (direct string HTML template,
    for short ones) or `template_name` (which points to the template to render).

    You can add fixed `context`, or override `get_plugin_context` which receives
    the context of the view the plugin is rendered in.

    In your template, use
    ```
    {% load gdaps %}
    <ul class="alerts">
    {% render_plugin IMyListItem %}
    </ul>
    ```
    in your template anywhere, and all of your implementations are rendered,
    one after each other, in a line. Each plugin can come from another app.
    You can change the order of the rendering using the `weight` attribute.
    """

    template: str = ""
    template_name: str = ""
    context: dict = {}
    weight = 0

    def get_plugin_context(self, context: Context) -> Context:
        """Override this method to add custom context to the plugin.

        :param context: the context where the plugin is rendered in.
            You can update it with own values, and return it.
            The return variable of this function will be the context
            of the rendered plugin. So if you don't update the passed
            context, but just return a new one, the plugin will not get
            access to the global context.

        Per default, it merges the plugin's ``context`` attribute into
        the given global context.
        """
        context.update(self.context)
        return context


class IFormExtensionMixin:
    """A mixin to create interfaces from, to extend Django forms."""
    class Media:
        js = ()
        css = {}

    field_names = []

    def alter_fields(self, fields: dict):
        """plugin hook for altering form fields.

        Attributes:
            fields: the form fields to alter. Can be changed in the method.
        """

    def alter_layout(self, layout):
        """plugin hook for altering a crispy form Layout() object.

        Call this method from your Form's  `__init__()` method.
        Attributes:
            layout: the crispy Layout object of the extended form.
                Modify it to your needs in this method.
        """

    def pre_save(self, instance, commit: bool) -> None:
        """Plugin hook to be called from a ModelForm's save() method before saving the instance."""

    def post_save(self, instance, commit: bool) -> None:
        """Plugin hook to be called from a ModelForm's save() method after saving the instance."""

    def clean(self, cleaned_data) -> dict:
        """Plugin hook to be called from within the extended Form's `clean()` method.

        Attributes:
            cleaned_data: the cleaned_data of the form's `clean()` method. Can be modified further.
        """


class IViewExtensionMixin:
    """A mixin to create Interfaces from, which extend Django views."""

    def alter_context_data(self, **kwargs) -> dict:
        """hook to alter context data of a view.

        Call this hook from get_context_data().
        """
        return {}

    def alter_response(self, request, response, *args, **kwargs) -> None:
        """hook to alter a response during a post/get/delete etc. request.

        Call this hook from get/post/delete().

        Attributes:
            request: the current request
            response: the response, as the post method created it so far.
                Feel free to modify it.
        """

    def form_valid(self, form, response) -> None:
        """hook that should be called by form_valid() to add plugin functionality."""



import logging

from pathlib import Path
from django.conf import settings
from django.core.management.base import CommandError, BaseCommand

from gdaps.conf import gdaps_settings

# from gdaps.pluginmanager import PluginManager

logger = logging.getLogger(__name__)

try:
    # if git is available, make use of git username/email config data
    import git

    reader = git.Repo.init(settings.BASE_DIR).config_reader()
except:
    reader = None
    logger.info(
        "If you want to take author/email automatically from current git repo, "
        "install gitpython using pip, and provide your data in the .gitconfig file."
    )


def get_user_data(key, default=""):
    if reader:
        return reader.get_value("user", key, default)
    else:
        return ""


class Command(BaseCommand):
    """This is the managemant command to add a plugin from a cookiecutter template to a Django application."""

    # absolute path to internal plugins of application
    # plugin_path = PluginManager.plugin_path()
    plugin_path = Path.cwd()
    help = (
        "Creates a basic GDAPS plugin from a cookiecutter template in the current working directory, e.g.\n"
        "You can also do a \n"
        "    cookiecutter gl:nerdocs/gdaps-plugin-cookiecutter"
    )
    missing_args_message = "You must provide a plugin name."

    def add_arguments(self, parser):
        parser.add_argument("name")

    def handle(self, name: str, **options):

        try:
            from cookiecutter.main import cookiecutter
        except ImportError:
            raise CommandError(
                "You have to install cookiecutter to make this command work."
            )

        if not name.isidentifier():
            raise CommandError(
                "The <name> parameter has to be a valid Python identifier."
            )

        logger.debug(f"Using plugin directory: {self.plugin_path}")

        try:
            target_path = Path(
                cookiecutter(
                    "gl:nerdocs/gdaps-plugin-cookiecutter",
                    extra_context={
                        "project_slug": settings.PROJECT_NAME,
                        "project_title": gdaps_settings.PROJECT_TITLE,
                        "plugin_title": name.replace("_", " ").capitalize(),
                        "app_name": name,
                        "author": get_user_data("name"),
                        "author_email": get_user_data("email"),
                    },
                    output_dir=self.plugin_path,
                )
            )
            # FIXME pip install -e only accepts relative paths (self.plugin_path)
            logger.info(
                f"\nSuccessfully created plugin: {target_path}\n"
                f"* Please edit '{target_path / 'setup.cfg'}' to your needs.\n"
                f"* Install the plugin locally using 'pip install -e {self.plugin_path}/{settings.PROJECT_NAME}-{name}'\n"
                f"  or with flit: 'flit install --symlink {self.plugin_path}/{settings.PROJECT_NAME}-{name}'\n"
                f"  Or you can add the plugin directly to your INSTALLED_APPS if you want it bundled with your application.\n"
                f"* Don't forget to call './manage.py syncplugins' after installing a new plugin to keep "
                f"your DB in sync.\n"
            )
        except Exception as e:
            logger.critical(f"There was an error with cookiecutter: \n {e}")



import os

import logging
import importlib
import warnings

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import QuerySet
from django.urls import include, path
from pkg_resources import iter_entry_points
from typing import List, Union

from gdaps.api import PluginConfig
from gdaps.exceptions import PluginError

__all__ = ["PluginManager"]

logger = logging.getLogger(__name__)


# with Python 3.7 we could use this instead of a plugin_spec dict:
#
# @dataclass
# class PluginSpec:
#     name: str
#     app_name: str
#     verbose_name: str
#     description: str
#     vendor: str
#     version: semver.VersionInfo
#     core_compat_version: semver.VersionInfo
#     author: str
#     author_email: str
#     category: str = "Misc"
#     enabled: bool = True
#     dependencies: list = ['core']


# class Singleton(type):
#     """A Metaclass implementing the Singleton pattern.
#
#     This class is for internal use only
#     """
#
#     _instances = {}
#
#     def __call__(cls, *args, **kwargs):
#         if cls not in cls._instances:
#             cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
#         return cls._instances[cls]


class PluginManager:
    """A Generic Django Plugin Manager that finds Django app plugins in a
    plugins folder or setuptools entry points and loads them dynamically.

    It provides a couple of methods to interaft with plugins, load submodules of all available plugins
    dynamically, or get a list of enabled plugins.
    Don't instantiate a ``PluginManager`` directly, just use its static and class methods directly.
    """

    group = ""
    _default_router = None

    def __init__(self):
        raise PluginError("PluginManager is not meant to be instantiated.")

    @classmethod
    def plugin_path(cls) -> str:
        """Returns the absolute path where application plugins live.

        This is basically the Django root + the dotted entry point.
        CAVE: this is not callable from within the settings.py file.
        """
        if not cls.group:
            raise ImproperlyConfigured(
                "Plugin path could not be determined. Please run PluginManager.find_plugins() in your settings.py first."
            )
        # TODO: use pathlib here
        return os.path.join(settings.BASE_DIR, *cls.group.split("."))

    @classmethod
    def find_plugins(cls, group: str) -> List[str]:
        """Finds plugins from setuptools entry points.

        This function is supposed to be called in settings.py after the
        INSTALLED_APPS variable. Therefore it can not use global variables from
        settings, to prevent circle imports.

        :param group: a dotted path where to find plugin apps. This is used as
            'group' for setuptools' entry points.
        :returns: A list of dotted app_names, which can be appended to
            INSTALLED_APPS.
        """

        warnings.warn(
            "The find_plugins() method is deprecated. Use alter_installed_apps() instead.",
            DeprecationWarning,
        )
        if not group:
            raise PluginError(
                "You have to specify an entry points group "
                "where GDAPS can look for plugins."
            )

        cls.group = group

        installed_plugin_apps = []
        for entry_point in iter_entry_points(group=group, name=None):
            appname = entry_point.module_name
            if entry_point.attrs:
                # FIXME: adding an AppConfig does not work yet
                appname += "." + ".".join(entry_point.attrs)

            installed_plugin_apps.append(appname)
            logger.info("Found plugin '{}'.".format(appname))

        return installed_plugin_apps

    @classmethod
    def alter_installed_apps(cls, installed_apps: list[str], group: str) -> list[str]:
        """Lets plugins update INSTALLED_APPS and add their own apps to it, in arbitrary order.

        Call this method directly after your settings.INSTALLED_APPS declaration.
        """
        if not group:
            raise PluginError(
                "You have to specify an entry points group "
                "where GDAPS can look for plugins."
            )

        cls.group = group

        for entry_point in iter_entry_points(group=group, name=None):
            appname = entry_point.module_name
            if entry_point.attrs:
                # FIXME: adding an AppConfig does not work yet
                module = importlib.import_module(appname)

                # if there is a function, call it
                if hasattr(module, "alter_installed_apps") and callable(
                    module.alter_installed_apps
                ):
                    module.alter_installed_apps(installed_apps)

                # if there is only a variable, append it
                elif hasattr(module, "INSTALLED_APPS"):
                    installed_apps += module.INSTALLED_APPS
                    appname += "." + ".".join(entry_point.attrs)
                    installed_apps.append(appname)
                    logger.info(
                        f"Found plugin '{appname}': {', '.join([app for app in module.INSTALLED_APPS])}"
                    )

                # legacy: if there is nothing, append the module name.
                else:
                    appname += "." + ".".join(entry_point.attrs)
                    if not appname in installed_apps:
                        installed_apps.append(appname)
                    logger.info(f"Found plugin '{appname}'.")

    @staticmethod
    def plugins(skip_disabled: bool = False) -> List[PluginConfig]:
        """Returns a list of AppConfig classes that are GDAPS plugins.

        This method basically checks for the presence of a ``PluginMeta`` attribute
        within the AppConfig of all apps and returns a list of apps containing it.
        :param skip_disabled: If True, skips disabled plugins and only returns enabled ones. Defaults to ``False``.
        """

        # TODO: test plugins() method
        list = []
        for app in apps.get_app_configs():
            if not hasattr(app, "PluginMeta"):
                continue
            if app.PluginMeta is None:
                continue
            if skip_disabled:
                # skip disabled plugins per default
                if not getattr(app.PluginMeta, "enabled", "True"):
                    continue
            list.append(app)

        return list

    @classmethod
    def load_plugin_submodule(cls, submodule: str, mandatory=False) -> list:
        """
        Search plugin apps for specific submodules and load them.

        :param submodule: the dotted name of the Django app's submodule to
            import. This package must be a submodule of the
            plugin's namespace, e.g. "schema" - then
            ["<main>.core.schema", "<main>.laboratory.schema"] etc. will be
            found and imported.
        :param mandatory: If set to True, each found plugin _must_ contain the given
            submodule. If any installed plugin doesn't have it, a PluginError is raised.
        :return: a list of module objects that have been successfully imported.
        """
        modules = []
        importlib.invalidate_caches()
        for app in PluginManager.plugins():

            # import all the submodules from all plugin apps

            dotted_name = f"{app.name}.{submodule}"
            try:
                module = importlib.import_module(dotted_name)
                modules.append(module)
                logger.info(f" ✓ Successfully loaded plugin submodule {dotted_name}")
            except ImportError as e:
                if mandatory:
                    raise PluginError(
                        f"The '{app.name}' app does not contain a (mandatory) '{submodule}' module"
                    )

                # if THIS module is not present, just ignore it
                if str(e) == f"No module named '{dotted_name}'":
                    logger.info(f" i Ignoring missing submodule '{dotted_name}'. {e}")
                else:
                    # else the importing has another reason, like syntax errors in that module, etc.
                    logger.error(f" ✘ Error importing submodule '{dotted_name}': {e}")
        return modules

    @classmethod
    def router(cls) -> Union["SimpleRouter", "DefaultRouter"]:
        """Loads all plugins' urls.py and collects their routers into one.

        :returns: a list of routers that can be merged with the global router."""

        try:
            from rest_framework.routers import DefaultRouter, SimpleRouter
        except ImportError:
            return []

        module_list = PluginManager.load_plugin_submodule("urls")

        if not cls._default_router:
            if settings.DEBUG:
                cls._default_router = DefaultRouter()
            else:
                cls._default_router = SimpleRouter()

            for module in module_list:
                router = getattr(module, "router", None)  # type: SimpleRouter
                if router:
                    logger.info(
                        f" ✓ Extended global router table with router from module '{module.__name__}'."
                    )
                    cls._default_router.registry.extend(router.registry)

        return cls._default_router

    @staticmethod
    def urlpatterns() -> list:
        """Loads all plugins' urls.py and collects their urlpatterns.

        This is maybe not the best approach, but it allows plugins to
        have "global" URLs, and not only namespaced, and it is flexible

        :returns: a list of urlpatterns that can be merged with the global
                  urls.urlpattern."""

        # FIXME: the order the plugins are loaded is not deterministic. This can lead to serious problems,
        # as apps could use the same URL namespace, and depending on which one was loaded first, it may mask the other
        # URL. This has to be fixed.
        #
        # Another unmanaged problem is 'dependencies':
        # FIXME: a dependency manager must be implemented into the PluginManager

        module_list = PluginManager.load_plugin_submodule("urls")

        urlpatterns = []
        for module in module_list:
            if module.__name__.startswith("gdaps"):
                continue

            root_urlpatterns = getattr(module, "root_urlpatterns", None)
            if root_urlpatterns:
                logger.info(
                    f" ✓ Added urlpatterns from module '{module.__name__}' to global list."
                )
                urlpatterns += root_urlpatterns
            try:
                namespace = getattr(module, "app_name")
            except AttributeError as e:
                raise ImproperlyConfigured(
                    "A GDAPS plugin's urls.py must define an 'app_name'. "
                    f"Please do that for '{module.__name__}'"
                )
            if getattr(module, "urlpatterns", None):
                urlpatterns += [path(namespace + "/", include(module))]

        return urlpatterns

    ###############################################################
    #  The following methods require Django's ORM already setup.  #
    #  Don't call them during the setup process                   #
    ###############################################################

    @staticmethod
    def orphaned_plugins() -> QuerySet:
        """Returns a list of GdapsPlugin models that have no disk representance any more.

        .. note:: This method needs Django's ORM to be running.
        """

        from gdaps.models import GdapsPlugin

        return GdapsPlugin.objects.exclude(
            name__in=[app.name for app in PluginManager.plugins()]
        )



import logging
import sys

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from gdaps import __version__
from gdaps.api import PluginConfig
from gdaps.pluginmanager import PluginManager

logger = logging.getLogger(__name__)


class GdapsPluginMeta:
    """This is the PluginMeta class of GDAPS itself."""

    version = __version__
    verbose_name = "Generic Django Application Plugin System"
    author = "Christian Gonzalez"
    author_email = "christian.gonzalez@nerdocs.at"
    category = "GDAPS"
    # hidden = True


class GdapsConfig(PluginConfig):
    name = "gdaps"
    PluginMeta = GdapsPluginMeta

    def ready(self):
        import pkg_resources

        # walk through all installed plugins and check some things
        for app in PluginManager.plugins():
            if hasattr(app.PluginMeta, "compatibility"):
                try:
                    pkg_resources.require(app.PluginMeta.compatibility)
                except pkg_resources.VersionConflict as e:
                    logger.critical("Incompatible plugins found!")
                    logger.critical(
                        f"Plugin {app.name} requires you to have {e.req}, but you installed {e.dist}."
                    )

                    sys.exit(1)

        if not settings.PROJECT_NAME:
            raise ImproperlyConfigured("GDAPS needs a PROJECT_NAME settings variable.")

        # load all generic gdaps.plugins - they must be implementations of GDAPS Interfaces
        logger.info("Loading gdaps plugins...")
        for entry_point in pkg_resources.iter_entry_points(
            group="gdaps.plugins", name=None
        ):  # type: pkg_resources.EntryPoint
            # it is enough to have them instantiated, as they are remembered internally in their interface.
            entry_point.load()



import os
import re
from django.core.exceptions import ValidationError

plugin_components_directory_name = "components"
plugin_models_directory_name = "models"
plugin_required_directories = [
    plugin_components_directory_name,
    plugin_models_directory_name,
]  # Todo: Check if they are provided


def validate_plugin_directory(file_path):
    file_path += "/" if file_path[-1] is not "/" else ""

    # check if all required directories exist
    for dir in plugin_required_directories:
        if dir != plugin_models_directory_name:  # models are not required
            path = "{}{}".format(file_path, dir)
            if not os.path.isdir(path):
                raise ValidationError(
                    "The required directory '{}' does not exist.".format(dir)
                )

    # check if components directory conains a .js or .jsx file
    component_files = os.listdir(
        "{}{}".format(file_path, plugin_components_directory_name)
    )
    if len([c for c in component_files if c[-3:] == ".js" or c[-4:] == ".jsx"]) == 0:
        raise ValidationError(
            "The component directory must contain at least one .js or .jsx file."
        )
    # Todo: Check if a .js or .jsx file with the name of the plugin exists and contains a default export React.Component


def validate_plugin_archive_fileextension(value):
    ext = os.path.splitext(value.name)[1]  # [0] returns path+filename
    valid_extensions = [".zip"]
    if not ext.lower() in valid_extensions:
        raise ValidationError("Unsupported file extension.")
